#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

"""Wrapper to configure mappings for a given tenant"""

from __future__ import annotations

import re
from copy import deepcopy
from json import loads

from cdk_proxy_api_client.common.logging import LOG
from cdk_proxy_api_client.errors import (
    GenericNotFound,
    ProxyApiException,
    ProxyGenericException,
)
from cdk_proxy_api_client.interceptors import Interceptors
from cdk_proxy_api_client.plugins import Plugins
from cdk_proxy_api_client.proxy_api import ProxyClient
from cdk_proxy_api_client.vclusters import VirturalClusters
from compose_x_common.compose_x_common import keyisset, set_else_none
from importlib_resources import files as pkg_files
from jsonschema import validate


def set_update_interceptors(proxy: ProxyClient, vcluster_name: str, interceptors: dict):
    plugins_client = Plugins(proxy)
    interceptors_client = Interceptors(proxy)
    try:
        plugins_list = plugins_client.list_all_plugins(as_list=True)
    except GenericNotFound as error:
        _plugins_list = plugins_client.list_all_plugins(as_list=True, extended=True)
        plugins_list = [p["pluginId"] for p in _plugins_list]

    for interceptor_name, interceptor_def in interceptors.items():
        plugin_class: str = interceptor_def.pop("pluginClass")
        if plugin_class not in plugins_list:
            raise ValueError(
                "The selected pluginClass {} is not available. Must be one of {}".format(
                    plugin_class, plugins_list
                )
            )
        if keyisset("username", interceptor_def):
            username = interceptor_def.pop("username")
        else:
            username = None
        try:
            interceptors_client.create_vcluster_interceptor(
                vcluster_name=vcluster_name,
                interceptor_name=interceptor_name,
                plugin_class=plugin_class,
                priority=int(interceptor_def["priority"]),
                config=interceptor_def["config"],
                username=username,
            )
            print(
                "Successfully created interceptor {} for {}".format(
                    interceptor_name, vcluster_name
                )
            )
        except ProxyGenericException:
            interceptors_client.update_vcluster_interceptor(
                vcluster_name=vcluster_name,
                interceptor_name=interceptor_name,
                plugin_class=plugin_class,
                priority=int(interceptor_def["priority"]),
                config=interceptor_def["config"],
                username=username,
            )
            print(f"Successfully updated {interceptor_name} for {vcluster_name}")
        except Exception as error:
            print(error)
            raise



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

import json
import logging
import os
import re
from copy import deepcopy
from datetime import datetime

import jwt
import yaml
from boto3 import Session
from cdk_proxy_api_client.client_wrapper import ApiClient
from cdk_proxy_api_client.proxy_api import ProxyClient
from cdk_proxy_api_client.vclusters import VirturalClusters
from compose_x_common.aws import get_assume_role_session
from compose_x_common.compose_x_common import keyisset

if logging.getLogger().hasHandlers():
    logging.getLogger().setLevel(logging.INFO)
else:
    logging.basicConfig(level=logging.INFO)


logger = logging.getLogger()

CDK_API_ENDPOINT = os.environ.get("CDK_API_ENDPOINT")
NEW_TOKEN_LIFETIME_IN_SECONDS = int(
    os.environ.get("NEW_TOKEN_LIFETIME_IN_SECONDS", 3660)
)
CENTRAL_FUNCTION_ARN_TO_INVOKE = os.environ.get("GW_ROTATION_MANAGER_FUNCTION_ARN")
if not CENTRAL_FUNCTION_ARN_TO_INVOKE:
    print(
        "No GW_ROTATION_MANAGER_FUNCTION_ARN set. Will attempt talking to GW directly."
    )
if not CENTRAL_FUNCTION_ARN_TO_INVOKE and not CDK_API_ENDPOINT:
    raise OSError(
        "CDK_API_ENDPOINT not set and no GW_ROTATION_MANAGER_FUNCTION_ARN to invoke."
        "One of the two must be set."
    )


def new_gateway_vcluster_secret_value(
    current_value: dict,
    lambda_session: Session,
) -> dict:
    """
    Calls the GW API Enpdoint to generate a new JWT token for the vcluster.
    It will then use the value of the AWSCURRENT secret and replace the content wit the new token.

    If CENTRAL_FUNCTION_ARN_TO_INVOKE is set, invokes that function to perform the token creation
    Else will call GW itself to get the new token

    If the current value is just a string, we replace it as-is.
    If the current value is a dict, we try upate the value of each key, if the old value is found in the dict value
    """
    current_jwt_token = current_value["SASL_PASSWORD"]
    current_username = current_value["SASL_USERNAME"]
    jwt_token_details: dict = get_vcluster_details_from_token(current_jwt_token)
    logger.info(
        "Token expiry was set to {}".format(
            datetime.utcfromtimestamp(jwt_token_details["exp"]).strftime("%FT%T")
        )
    )
    if CENTRAL_FUNCTION_ARN_TO_INVOKE:
        logger.info("Using Lambda function to get new token.")
        new_jwt_token = get_token_from_central_function(
            lambda_session, jwt_token_details
        )
    else:
        gateway_secret = get_cdk_gw_admin_creds(lambda_session)
        logger.info("Successfully retrieved GW admin credentials")
        try:
            logger.info(
                f"creating token for vcluster - {jwt_token_details['vcluster']} - {jwt_token_details['username']}"
            )

            new_jwt_token = get_new_token_for_vcluster(
                gateway_secret, jwt_token_details, NEW_TOKEN_LIFETIME_IN_SECONDS
            )
        except Exception as error:
            logger.exception(error)
            logger.error(
                "Failed to create a new JWT token for vcluster {}".format(
                    jwt_token_details["vcluster"]
                )
            )
            raise
    new_secret_value = replace_string_in_dict_values(
        current_value,
        current_username,
        jwt_token_details["username"],
        True,
    )
    replace_string_in_dict_values(new_secret_value, current_jwt_token, new_jwt_token)
    return new_secret_value


def get_token_from_central_function(
    lambda_session: Session, jwt_token_details: dict
) -> str:
    """Function that will invoke sync another Function in charge of getting the new token from a central function"""
    cdk_api_secret_arn = os.environ.get("CDK_API_SECRET_ARN_ROLE")
    client = set_calls_clients(cdk_api_secret_arn, lambda_session).client("lambda")
    try:
        new_token_r = client.invoke(
            FunctionName=CENTRAL_FUNCTION_ARN_TO_INVOKE,
            Payload=json.dumps(
                {
                    "vcluster": jwt_token_details["vcluster"],
                    "username": jwt_token_details["username"],
                    "expiry": NEW_TOKEN_LIFETIME_IN_SECONDS,
                }
            ),
        )
        res = json.loads(new_token_r["Payload"].read())
    except Exception as error:
        print("Remote function invoke failure", error)
        raise
    return res["token"]


def set_calls_clients(cdk_api_secret_arn: str, lambda_session: Session) -> Session:
    cdk_api_secret_role_arn = os.environ.get("CDK_API_SECRET_ARN_ROLE", None)
    if cdk_api_secret_role_arn and cdk_api_secret_arn.startswith("arn:aws"):
        return get_assume_role_session(lambda_session, cdk_api_secret_role_arn)
    return lambda_session


def get_cdk_gw_admin_creds(lambda_session: Session) -> dict:
    cdk_api_secret_arn = os.environ.get("CDK_API_SECRET_ARN")
    if not cdk_api_secret_arn:
        raise OSError("CDK_API_SECRET_ARN must be set.")
    client = set_calls_clients(cdk_api_secret_arn, lambda_session).client(
        "secretsmanager"
    )
    try:
        secret_value = client.get_secret_value(SecretId=cdk_api_secret_arn)[
            "SecretString"
        ]
    except Exception as error:
        logger.exception(error)
        logger.error(f"Failed to retrieve the SecretString for {cdk_api_secret_arn}")
        raise

    try:
        gw_secret = yaml.safe_load(secret_value)
    except yaml.YAMLError as error:
        logger.exception(error)
        logger.info("Secret string is not YAML formatted")
        try:
            gw_secret = json.loads(secret_value)
        except json.JSONDecodeError as error:
            logger.exception(error)
            logger.info("Secret string is not JSON formatted.")
            gw_secret = secret_value

    if not isinstance(gw_secret, (list, dict)):
        raise TypeError(
            "The secret format is not valid. Got {}, expected one of {}".format(
                str, (list, dict)
            )
        )
    for secret in gw_secret:
        if keyisset("admin", secret):
            return secret
    raise ValueError(f"No admin user found in {cdk_api_secret_arn}")


def get_vcluster_details_from_token(jwt_token: str) -> dict:
    """Uses existing secret JWT token to identify vcluster owner"""
    try:
        jwt_content = jwt.decode(
            jwt_token,
            options={"verify_signature": False},
            algorithms=["HS256", "HS384", "HS512"],
        )
        return jwt_content
    except Exception as error:
        logger.error(f"Jwt token decode error + {error}")
        raise error


def get_new_token_for_vcluster(
    gw_admin_secret: dict, vcluster: dict, life, token_only: bool = True
) -> str:
    try:
        logger.info(
            f"creating secret for {vcluster['vcluster']}/{vcluster['username']} with user on host {CDK_API_ENDPOINT}"
        )
        api_client = ApiClient(
            username=gw_admin_secret["username"],
            password=gw_admin_secret["password"],
            url=CDK_API_ENDPOINT,
        )
        proxy_client = ProxyClient(api_client)
        admin_client = VirturalClusters(proxy_client)
        logger.debug(admin_client.list_vclusters(as_list=True))
        token = admin_client.create_vcluster_user_token(
            vcluster=vcluster["vcluster"],
            username=vcluster["username"],
            lifetime_in_seconds=life,
            token_only=token_only,
        )
        return token
    except Exception as error:
        logger.exception(error)
        logger.error(
            "get_new_token_for_vcluster : Failed to create new jwt token for {}".format(
                vcluster["vcluster"]
            )
        )


def replace_string_in_dict_values(
    input_dict: dict, src_value: str, new_value: str, copy: bool = False
) -> dict:
    """
    Function to update all values in an input dictionary that will match with the new value.
    It uses the new value to set as regex and replaces it in the dict values for each key,
    allowing for any secret value and dynamic dict content.
    """

    src_re = re.compile(re.escape(src_value))
    if copy:
        updated_dict = deepcopy(input_dict)
    else:
        updated_dict = input_dict
    for key, value in updated_dict.items():
        if isinstance(value, str) and src_re.findall(value):
            updated_dict[key] = src_re.sub(new_value, value)
        elif not isinstance(value, str):
            print(f"The value for {key} is not a string. Skipping")
    if copy:
        return updated_dict



# SPDX-License-Identifier: Apache-2.0
# Copyright 2023 John Mille <john@ews-network.net>

from __future__ import annotations

from argparse import ArgumentParser

VCLUSTER_PARSER = ArgumentParser(add_help=False)
VCLUSTER_PARSER.add_argument(
    "--vcluster-name",
    dest="vcluster_name",
    required=True,
    help="Name of the vcluster to make operations for",
)

USERNAME_PARSER = ArgumentParser(add_help=False)
USERNAME_PARSER.add_argument(
    "--username",
    dest="vcluster_username",
    required=False,
    help="Specify username to get interceptors for",
)


def set_interceptors_actions_parsers(vclusters_subparsers):
    """Creates all the parser and subparsers for vClusters API endpoints"""

    interceptors_parser = vclusters_subparsers.add_parser(
        name="interceptors",
        help="List vCluster interceptors",
        parents=[VCLUSTER_PARSER],
    )
    interceptors_subparsers = interceptors_parser.add_subparsers(dest="sub_action")
    list_parser = interceptors_subparsers.add_parser(
        name="list",
        help="List interceptors for vcluster/username",
        parents=[USERNAME_PARSER],
    )
    create_parser = interceptors_subparsers.add_parser(
        name="create-update",
        help="Create or update a new vCluster mapping",
        parents=[USERNAME_PARSER],
    )
    # create_parser.add_argument("--interceptor-name", type=str, required=True)
    create_parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to config (YAML/JSON) to use for interceptor",
    )

    # Delete action parsers
    delete_interceptor_mapping_parser = interceptors_subparsers.add_parser(
        name="delete",
        help="Delete interceptor",
        parents=[USERNAME_PARSER],
    )
    delete_interceptor_mapping_parser.add_argument(
        "--interceptor-name",
        dest="interceptor_name",
        required=True,
        help="Interceptor name to delete as seen in the vCluster.",
    )
    import_from_config_parser = interceptors_subparsers.add_parser(
        name="import-from-config",
        help="Import interceptors from config file",
        parents=[USERNAME_PARSER],
    )
    import_from_config_parser.add_argument(
        "-f",
        "--config-file",
        dest="InputConfigFile",
        type=str,
        help="Path to the config file (YAML)",
        required=True,
    )



#   SPDX-License-Identifier: Apache-2.0
#   Copyright 2023 John Mille <john@ews-network.net>

"""Wrapper to configure mappings for a given tenant"""

from __future__ import annotations

import re
from copy import deepcopy
from json import loads

from cdk_proxy_api_client.common.logging import LOG
from cdk_proxy_api_client.errors import ProxyApiException, ProxyGenericException
from cdk_proxy_api_client.proxy_api import ProxyClient
from cdk_proxy_api_client.vclusters import VirturalClusters
from compose_x_common.compose_x_common import keyisset, set_else_none
from importlib_resources import files as pkg_files
from jsonschema import validate

DEFAULT_SCHEMA_PATH = pkg_files("cdk_gw_tools").joinpath(
    "specs/tenant_mappings-input.json"
)


def get_tenant_logical_topics(
    proxy: ProxyClient, tenant_name: str, include_read_only: bool = False
) -> list[dict]:
    """
    Returns the list of topics (logical name) from a given tenant topics mappings. Ignores
    :param ProxyClient proxy:
    :param str tenant_name:
    :param bool include_read_only:
    :return: List of topics logical available in the tenant.
    """
    tenant_mappings = VirturalClusters(proxy)

    topics_list: list[dict] = []
    for _topic in tenant_mappings.list_vcluster_topic_mappings(
        tenant_name, as_list=True
    ):
        if _topic["readOnly"] is True and not include_read_only:
            continue
        topics_list.append(_topic)
    return topics_list


def import_from_tenants_include_string(
    proxy: ProxyClient,
    include_regex: str,
    vcluster_name: str,
    tenants: list[str],
    processed_tenants: list[str],
    process_once: bool = False,
) -> None:
    """Matches tenant based on simple string regex"""
    try:
        _pattern = re.compile(include_regex)
    except Exception as error:
        print(error)
        print(include_regex, "Not a valid regex")
        return
    tenant_mappings = VirturalClusters(proxy)
    for _tenant in tenants:
        if _pattern.match(_tenant):
            if process_once and _tenant in processed_tenants:
                print(
                    f"Tenant {_tenant} was already processed. Skipping",
                    processed_tenants,
                )
                continue
            else:
                tenant_topics: list[dict] = get_tenant_logical_topics(proxy, _tenant)
                for _import_tenant_topic in tenant_topics:
                    tenant_mappings.create_vcluster_topic_mapping(
                        vcluster_name,
                        _import_tenant_topic["logicalTopicName"],
                        _import_tenant_topic["physicalTopicName"],
                        read_only=False,
                    )
                processed_tenants.append(_tenant)


def import_from_tenants_include_dict(
    proxy: ProxyClient,
    mapping_import_config: dict,
    tenant_name: str,
    tenants: list[str],
    processed_tenants: list[str],
    process_once: bool = False,
) -> None:
    """Import topic mappings tenants from complex definition"""
    try:
        _pattern = re.compile(mapping_import_config["tenant_regex"])
    except Exception as error:
        print(error)
        print(mapping_import_config["tenant_regex"], "Not a valid regex")
        return
    topics_exclude_pattern_regexes: list = set_else_none(
        "logical_topics_exclude_regexes", mapping_import_config
    )
    topics_include_pattern_regexes: list = set_else_none(
        "logical_topics_include_regexes", mapping_import_config, ["^(.*)$"]
    )
    topics_exclude_patterns: list[re.Pattern] = []
    topics_include_patterns: list[re.Pattern] = []
    if topics_exclude_pattern_regexes:
        for _exclude_pattern in topics_exclude_pattern_regexes:
            try:
                topics_exclude_patterns.append(re.compile(_exclude_pattern))
            except Exception as error:
                LOG.exception(error)
                LOG.error(
                    "logical_topics_exclude_regexes",
                    _exclude_pattern,
                    "Not a valid regex. Skipping",
                )
                return
    if topics_include_pattern_regexes:
        for _include_pattern in topics_include_pattern_regexes:
            try:
                topics_include_patterns.append(re.compile(_include_pattern))
            except Exception as error:
                LOG.exception(error)
                LOG.error(
                    "logical_topics_include_regexes",
                    _include_pattern,
                    "Not a valid regex. Skipping",
                )
                return

    tenant_mappings = VirturalClusters(proxy)
    for _tenant in tenants:
        if _pattern.match(_tenant):
            if process_once and _tenant in processed_tenants:
                LOG.debug(
                    "Tenant {} was already processed. Skipping {}".format(
                        _tenant, processed_tenants
                    ),
                )
                continue
            else:
                processed_tenants.append(_tenant)
        else:
            LOG.debug(f"Skipping tenant {_tenant}")
            continue
        tenant_topics: list[dict] = get_tenant_logical_topics(proxy, _tenant)
        topics_to_import: list[dict] = deepcopy(tenant_topics)
        for _import_tenant_topic in tenant_topics:
            for _exclude_pattern in topics_exclude_patterns:
                if (
                    _exclude_pattern.match(_import_tenant_topic["logicalTopicName"])
                    and _import_tenant_topic["logicalTopicName"] in topics_to_import
                ):
                    topics_to_import.remove(_import_tenant_topic)
                    LOG.debug(
                        "Topic: {} matched against exclude regex: {}".format(
                            _import_tenant_topic["logicalTopicName"],
                            _exclude_pattern.pattern,
                        ),
                    )
        LOG.debug(
            "Topics post exclude", [_t["logicalTopicName"] for _t in topics_to_import]
        )
        final_topics_import: list[dict] = []
        for _import_tenant_topic in topics_to_import:
            for _include_pattern in topics_include_patterns:
                if (
                    _include_pattern.match(_import_tenant_topic["logicalTopicName"])
                    and _import_tenant_topic not in final_topics_import
                ):
                    final_topics_import.append(_import_tenant_topic)
                    break
        LOG.debug(f"Tenant:{_tenant} - Final topic list:{final_topics_import}")
        grant_write_access = keyisset("grant_write_access", mapping_import_config)
        for topic_mapping in final_topics_import:
            try:
                tenant_mappings.create_vcluster_topic_mapping(
                    tenant_name,
                    topic_mapping["logicalTopicName"],
                    topic_mapping["physicalTopicName"],
                    read_only=not grant_write_access,
                )
                if grant_write_access:
                    LOG.warn(
                        "{}:{} - Granting write access.".format(
                            _tenant, topic_mapping["physicalTopicName"]
                        )
                    )
            except ProxyApiException:
                print(
                    "Failed to create mapping from {}: {} does not seem to exist.".format(
                        _tenant, topic_mapping["physicalTopicName"]
                    )
                )


def import_from_other_tenants(
    proxy: ProxyClient, import_config: dict, tenant_name: str
) -> None:
    """Allows to import existing topics from other tenants in read-only"""
    tenants: list[str] = VirturalClusters(proxy).list_vclusters(as_list=True)
    exclude_list = set_else_none("exclude_regex", import_config, [])
    include_list = set_else_none("include_regex", import_config, [])
    processed_tenants: list[str] = []
    process_once: bool = keyisset("process_tenant_only_once", import_config)
    if not include_list:
        raise ValueError("There must be at least one item in include_regex")
    for _regex in exclude_list:
        try:
            _pattern = re.compile(_regex)
            for _tenant in tenants:
                if _pattern.match(_tenant):
                    tenants.remove(_tenant)
        except Exception as error:
            print(error)
            print(_regex, "not a valid regex. Ignoring")
    for _include_item in include_list:
        if isinstance(_include_item, str):
            import_from_tenants_include_string(
                proxy,
                _include_item,
                tenant_name,
                tenants,
                processed_tenants,
                process_once,
            )
        elif isinstance(_include_item, dict):
            import_from_tenants_include_dict(
                proxy,
                _include_item,
                tenant_name,
                tenants,
                processed_tenants,
                process_once,
            )
        else:
            raise TypeError(
                _include_item,
                "is of type",
                type(_include_item),
                "expected one of",
                (str, dict),
            )


def propagate_tenant_mappings(
    tenant_mappings: VirturalClusters,
    mappings: list[dict],
    tenant_name: str,
    ignore_conflicts: bool = False,
) -> None:
    for mapping in mappings:
        print("MAPPINGS IMPORT", mapping)
        try:
            tenant_mappings.create_vcluster_topic_mapping(
                tenant_name,
                mapping["logicalTopicName"],
                mapping["physicalTopicName"],
                read_only=keyisset("readOnly", mapping),
                concentrated=keyisset("concentrated", mapping),
            )
        except ProxyGenericException as error:
            if error.code == 409 and ignore_conflicts:
                pass


def import_tenants_mappings(
    client: ProxyClient, config_content: dict, tenant_name: str, schema: dict = None
) -> list[dict]:
    """Will create mappings from the config content, and return the final mappings for the tenant"""
    if not schema:
        schema = loads(DEFAULT_SCHEMA_PATH.read_text())
    validate(config_content, schema)
    tenant_name = set_else_none("tenant_name", config_content, tenant_name)
    ignore_conflicts = keyisset("ignore_duplicates_conflict", config_content)
    mappings = config_content["mappings"]
    tenant_mappings = VirturalClusters(client)
    propagate_tenant_mappings(tenant_mappings, mappings, tenant_name, ignore_conflicts)
    import_from_other_tenants_config = set_else_none(
        "import_from_tenant", config_content
    )
    if import_from_other_tenants_config:
        import_from_other_tenants(client, import_from_other_tenants_config, tenant_name)

    return tenant_mappings.list_vcluster_topic_mappings(tenant_name, True)
